;;
;; x-face-el-init.el
;;
;; Takuo KITAME <kitame@northeye.org>

(setq load-path 
      (cons 
       (concat "/usr/share/" (symbol-name debian-emacs-flavor)
			   "/site-lisp/x-face-el")
       load-path))
(setq x-face-inhibit-loadup-splash t)
(cond ((featurep 'xemacs)
       ;; settings for XEmacs.
	   (autoload 'x-face-encode "x-face"
		 "Generate X-Face string(s) from xbm file." t)
	   (autoload 'x-face-insert "x-face"
		 "Insert X-Face fields." t)
	   (autoload 'x-face-save "x-face"
		 "Save X-Face fields to files." t)
	   (autoload 'x-face-view "x-face"
		 "View X-Face fields." t)
	   (autoload 'x-face-ascii-view "x-face"
		 "View X-Face fields as ascii pictures." t)
	   (autoload 'x-face-xmas-display-x-face "x-face"
		 "Display X-Face fields as XEmacs-glyph." t)
	   (autoload 'x-face-xmas-force-display-x-face "x-face"
		 "Display X-Face fields compulsorily as XEmacs glyph." t)
	   (autoload 'x-face-xmas-remove-x-face-glyph "x-face"
		 "Remove X-Face images and some text-properties." t)
	   (require 'x-face-xmas)
	   )
      ((and (boundp 'emacs-major-version) (eq emacs-major-version 21))
       ;;
       ;; You may have no need to use the following one line if you
       ;; always use T-gnus 6.14.5 revision 07 and later.
       (autoload 'x-face-decode-message-header "x-face-e21")       ;;
       (autoload 'x-face-insert "x-face-e21" nil t)
       (autoload 'x-face-save "x-face-e21" nil t)
       (autoload 'x-face-show "x-face-e21" nil t)
       (autoload 'x-face-turn-off "x-face-e21")
	   (autoload 'x-face-ascii-view "x-face" nil t)
       ;;
       ;; Show X-Face images when `x-face-insert' is done.
       (setq x-face-auto-image t)
       ;;
       ;; If you show X-Face images in the message sending buffer,
       ;; it is STRONGLY recommended that you remove images from the
       ;; buffer before sending a message.  However, it seems not to
       ;; be required for Gnusae so far.  The following lines are for
       ;; SEMI and Mew.  The latter can be put into .mew file instead.
       (add-hook 'mime-edit-translate-hook 'x-face-turn-off)
       (add-hook 'mew-make-message-hook 'x-face-turn-off)
	   )
      (t
       ;; settings for the other Emacsen.
	   (autoload 'x-face-encode "x-face"
		 "Generate X-Face string(s) from xbm file." t)
	   (autoload 'x-face-insert "x-face"
		 "Insert X-Face fields." t)
	   (autoload 'x-face-save "x-face"
		 "Save X-Face fields to files." t)
	   (autoload 'x-face-view "x-face"
		 "View X-Face fields." t)
	   (autoload 'x-face-ascii-view "x-face"
		 "View X-Face fields as ascii pictures." t)

	   (setq x-face-save-or-view-method-alist
			 '(
			   (gnus-summary-mode
				. (gnus-summary-show-article
				   x-face-find-gnus-article-buffer
				   gnus-summary-show-article))
			   (gnus-Subject-mode
				. (gnus-Subject-show-article
				   x-face-find-gnus-article-buffer
				   gnus-Subject-show-article))
			   (wl-summary-mode
				. (x-face-wl-get-article
				   x-face-extracted-headers-buffer))
			   (cmail-summary-mode
				. (x-face-cmail-get-article
				   x-face-extracted-headers-buffer))
			   (mh-folder-mode
				. (x-face-mh-get-article
				   x-face-extracted-headers-buffer))
			   (vm-summary-mode
				. (x-face-vm-get-all-headers
				   x-face-extracted-headers-buffer))
			   (mew-summary-mode
				. (x-face-mew-get-article
				   x-face-extracted-headers-buffer))
			   (mew-virtual-mode
				. (x-face-mew-get-article
				   x-face-extracted-headers-buffer))
			   (require 'x-face)
			   ))
       ))

(setq x-face-compressor-program-list
      '((gzip ".gz" "gzip" "-c9" "GZIPping" "gzip" "-cd" "GUNZIPping")
		(bzip2 ".bz2" "bzip2" "-f" "BZIPping" "bzip2" "-d" "BUNZIPping")))

(setq x-face-compface-program "compface")
(setq x-face-uncompface-program "uncompface")

(mapcar
 (function (lambda (hook)
			 (add-hook hook
					   '(lambda () 
						  (cond ((featurep 'xemacs) 
								 (x-face-xmas-replace-highlight-headers)
								 (x-face-xmas-replace-x-face-function-for-semi))
								((and (boundp 'emacs-major-version) (eq emacs-major-version 21))
								 ;;
								 )
								(t
								 ;;
								 ))
						  ))
			 ))
 '(gnus-message-setup-hook 
   gnus-summary-mode-hook
   wl-summary-mode-hook
   wl-mail-setup-hook
   cmail-summary-mode-hook
   mail-setup-hook
   mh-folder-mode-hook
   mh-letter-mode-hook
   rmail-mode-hook
   vm-summary-mode-hook
   mew-draft-mode-hook
   mew-summary-mode-hook
   mew-virtual-mode-hook))

(cond ((featurep 'xemacs)
	   ;; inline display in draft
	   (setq x-face-xmas-auto-image t)
	   ;; for gnus
	   (setq gnus-article-display-hook nil)
	   ;; for Wanderlust
	   (autoload 'x-face-xmas-wl-display-x-face "x-face" nil t)
	   (if (boundp 'wl-icon-directory)
		   (setq wl-highlight-x-face-function 'x-face-xmas-wl-display-x-face)
		(setq wl-highlight-x-face-func 'x-face-xmas-wl-display-x-face))
	   ;; for Mew
	   (autoload 'x-face-xmas-mew-display-x-face "x-face" nil t)
	   (setq mew-use-highlight-x-face-function 'x-face-xmas-mew-display-x-face)
	  ;; for SEMI
	   (add-hook 'mime-edit-translate-hook
				 'x-face-xmas-remove-x-face-glyph)
	   )
	  ((or (and (boundp 'emacs-major-version) (eq emacs-major-version 21))
		   (featurep 'x-face-mule))
	   ;; view on MUAs
	   (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
	   (setq gnus-article-x-face-command 'x-face-decode-message-header
			 gnus-treat-display-xface 'head)
	   (setq mew-use-highlight-x-face t
			 mew-use-highlight-x-face-function 'x-face-decode-message-header)
	   (add-hook 'mh-show-mode-hook 'x-face-decode-message-header)
	   (add-hook 'rmail-show-message-hook 'x-face-decode-message-header)
	   (if (boundp 'wl-icon-directory)
		   (setq wl-highlight-x-face-function 'x-face-decode-message-header)
		 (setq wl-highlight-x-face-func 'x-face-decode-message-header))
	   ))

;;;***

;;; Local Variables:
;;; mode: emacs-lisp
;;; tab-width: 4
;;; End:
