;;; x-face-e21.el --- X-Face utilities for Emacs 21

;; Copyright (C) 2000, 2001, 2002 Free Software Foundation, Inc.

;; Author: Katsumi Yamaoka  <yamaoka@jpl.org>
;; Created: 2000/06/16
;; Revised: 2002/02/20
;; Keywords: x-face, cmail, emh, gnus, mew, mh-e, rmail, vm, wanderlust, bbdb

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; At first, you have to install this file manually to the location
;; which is referred by `load-path'.  For example:
;;
;; # cp -p x-face-e21.el /usr/local/share/emacs/site-lisp/
;; # cd /usr/local/share/emacs/site-lisp/
;; # emacs -batch -f batch-byte-compile x-face-e21.el

;; If you would like to use BBDB, it should be the version 2.2 or
;; later.  The old versions are out of support even if it works.

;; If you wish to scale down or scale up X-Face images, the external
;; commands "pnmscale" and "pgmtoppm" are required.  This feature has
;; been tested with netpbm-9.24 and its bundled commands.

;; Gray X-Faces are now supported.  So far the feature to save those
;; X-Face headers in XBM files has not been implemented yet.

;; The command `x-face-insert' allows not only XBM files but also pre-
;; encoded data files.  Therefore, if you already have such a file,
;; for example, ~/.xface which is the default value of
;; `wl-x-face-file' or `mew-x-face-file', you can use it just as it
;; is.

;; Here are some examples of how to use this package with various MUAs
;; and BBDB.  If you are using Emacs versions prior to 19.23 besides
;; Emacs 21, replace the line
;;      ((eq emacs-major-version 21)
;; with
;;      ((and (boundp 'emacs-major-version) (eq emacs-major-version 21))

;;[Common settings for all MUAs] For your .emacs file:
;;----------------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;; settings for XEmacs.
;;       )
;;      ((eq emacs-major-version 21)
;;       ;;
;;       ;; You may have no need to use the following one line if you
;;       ;; always use T-gnus 6.14.5 revision 07 and later.
;;       (autoload 'x-face-decode-message-header "x-face-e21")
;;       ;;
;;       (autoload 'x-face-insert "x-face-e21" nil t)
;;       (autoload 'x-face-save "x-face-e21" nil t)
;;       (autoload 'x-face-show "x-face-e21" nil t)
;;       (autoload 'x-face-turn-off "x-face-e21")
;;       ;;
;;       ;; Show X-Face images when `x-face-insert' is done.
;;       (setq x-face-auto-image t)
;;       ;;
;;       ;; If you show X-Face images in the message sending buffer,
;;       ;; it is STRONGLY recommended that you remove images from the
;;       ;; buffer before sending a message.  However, it seems not to
;;       ;; be required for Gnusae so far.  The following lines are for
;;       ;; SEMI and Mew.  The latter can be put into .mew file instead.
;;       (add-hook 'mime-edit-translate-hook 'x-face-turn-off)
;;       (add-hook 'mew-make-message-hook 'x-face-turn-off))
;;      (t
;;       ;; settings for the other Emacsen.
;;       ))

;;[Cmail with SEMI] For your .emacs file:
;;---------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "cmail-modes"
;;	 '(progn
;;	    (define-key cmail-summary-mode-map "\C-x4s" 'x-face-save)
;;	    (define-key cmail-mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key cmail-mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'cmail-mail-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[Gnusae] For your .gnus file:
;;-----------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       ;;
;;       ;; You may have no need to use the following two lines if you
;;       ;; always use T-gnus 6.14.5 revision 07 and later.
;;       (setq gnus-article-x-face-command 'x-face-decode-message-header
;;	     gnus-treat-display-xface 'head)
;;       ;;
;;       (eval-after-load "gnus-sum"
;;	 '(define-key gnus-summary-mode-map "\C-x4s" 'x-face-save))
;;       (define-key message-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key message-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'gnus-message-setup-hook 'x-face-insert)
;;       ;;
;;       ;; Uncomment the following line if you would like to show gray
;;       ;; X-Faces using this program instead of the native function.
;;       ;;
;;       ;;(setq gnus-treat-display-grey-xface nil)
;;       )
;;      (t
;;       ;;
;;       ))

;;[Mew] For your .mew file:
;;-------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (setq mew-use-highlight-x-face t
;;	     mew-use-highlight-x-face-function 'x-face-decode-message-header)
;;       (define-key mew-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key mew-draft-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key mew-draft-header-map "\C-x4i" 'x-face-insert)
;;       (define-key mew-draft-mode-map "\M-t" 'x-face-show)
;;       (define-key mew-draft-header-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mew-draft-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[MH-E without EMH] For your .emacs file:
;;----------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'mh-show-mode-hook 'x-face-decode-message-header)
;;       (eval-after-load "mh-e"
;;	 '(define-key mh-folder-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "mh-comp"
;;	 '(progn
;;	    (define-key mh-letter-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mh-letter-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mh-letter-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[MH-E with EMH] For your .emacs file:
;;-------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "mh-e"
;;	 '(define-key mh-folder-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "mh-comp"
;;	 '(progn
;;	    (define-key mh-letter-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mh-letter-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mh-letter-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[RMAIL without RMAIL-MIME] For your .emacs file:
;;------------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'rmail-show-message-hook 'x-face-decode-message-header)
;;       (eval-after-load "rmail"
;;	 '(define-key rmail-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "rmailsum"
;;	 '(define-key rmail-summary-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "sendmail"
;;	 '(progn
;;	    (define-key mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mail-setup-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[RMAIL with RMAIL-MIME] For your .emacs file:
;;---------------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (eval-after-load "rmail"
;;	 '(define-key rmail-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "rmailsum"
;;	 '(define-key rmail-summary-mode-map "\C-x4s" 'x-face-save))
;;       (eval-after-load "sendmail"
;;	 '(progn
;;	    (define-key mail-mode-map "\C-x4i" 'x-face-insert)
;;	    (define-key mail-mode-map "\M-t" 'x-face-show)))
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'mail-setup-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[VM without SEMI] For your .vm file:
;;------------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (eval-after-load "vm-page"
;;	 '(defadvice vm-energize-headers-and-xfaces
;;	    (after x-face-decode-message-header activate compile)
;;	    "Show X-Face using `x-face-decode-message-header'."
;;	    (save-restriction
;;	      (widen)
;;	      (narrow-to-region (vm-headers-of (car vm-message-pointer))
;;				(vm-text-of (car vm-message-pointer)))
;;	      (x-face-decode-message-header))))
;;       (define-key vm-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key vm-mail-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key vm-mail-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'vm-mail-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[VM with SEMI] For your .vm file:
;;---------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (add-hook 'mime-display-header-hook 'x-face-decode-message-header)
;;       (define-key vm-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key vm-mail-mode-map "\C-x4i" 'x-face-insert)
;;       (define-key vm-mail-mode-map "\M-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'vm-mail-mode-hook 'x-face-insert))
;;      (t
;;       ;;
;;       ))

;;[Wanderlust] For your .wl file:
;;-------------------------------
;;(cond ((featurep 'xemacs)
;;       ;;
;;       )
;;      ((eq emacs-major-version 21)
;;       (setq wl-highlight-x-face-function 'x-face-decode-message-header)
;;       (define-key wl-summary-mode-map "\C-x4s" 'x-face-save)
;;       (define-key wl-draft-mode-map "\C-x4i" 'x-face-insert)
;;       ;; "\M-t" key is reserved for wl command.
;;       (define-key wl-draft-mode-map "\M-\C-t" 'x-face-show)
;;       ;;
;;       ;; If a file name has no directory component, it should be
;;       ;; found in the directory which is specified by the option
;;       ;; `x-face-image-file-directory'.
;;       (setq x-face-default-xbm-file "YourFace.xbm")
;;       ;;
;;       (add-hook 'wl-mail-setup-hook 'x-face-insert)
;;       ;; If you use `wl-draft-insert-x-face-field' instead of
;;       ;; `x-face-insert' for inserting an X-Face, you can highlight
;;       ;; it as an image with the setting of the following hook:
;;       (add-hook 'wl-draft-insert-x-face-field-hook
;;		 (lambda nil
;;		   (x-face-insert wl-x-face-file)))
;;       )
;;      (t
;;       ;;
;;       ))

;;[BBDB]
;;------
;; To make BBDB automatically add X-Face fields in the record, put the
;; following lines in your startup (commonly .emacs) file after setting
;; up the value of `bbdb-auto-notes-alist' or after the library
;; "bbdb-hooks" is loaded.
;;
;;(put 'face 'field-separator "\n")
;;(setq bbdb-auto-notes-alist
;;      (append bbdb-auto-notes-alist
;;	      `(("x-face"
;;		 (,(concat "[\t\n ]*\\([^\t\n ]*\\)"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?"
;;			   "\\([\t\n ]+\\([^\t\n ]+\\)\\)?")
;;		  face
;;		  "\\1\\3\\5\\7\\9")))))
;;
;; For only Emacs 21, you may use the following instead:
;;
;;(put 'face 'field-separator "\n")
;;(setq bbdb-auto-notes-alist
;;      (append bbdb-auto-notes-alist
;;	      `(("x-face"
;;		 (,(concat "[[:blank:]\n]*\\([^[:blank:]\n]*\\)"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?"
;;			   "\\(?:[[:blank:]\n]+\\([^[:blank:]\n]+\\)\\)?")
;;		  face
;;		  "\\1\\2\\3\\4\\5")))))

;;; Code:

(eval-and-compile
  (if (or (featurep 'xemacs)
	  (not (boundp 'emacs-major-version))
	  (< emacs-major-version 21))
      (error "This program is for FSF Emacs 21+, good-bye")))

(defconst x-face-e21-version "0.112")

(defgroup x-face nil
  "X-Face utilities."
  :group 'news
  :group 'mail)

(defcustom x-face-image-file-directory "~/x-faces/"
  "Name of the directory where the image files can be found."
  :version "21.1"
  :type 'directory
  :group 'x-face)

(defcustom x-face-default-xbm-file nil
  "Default xbm file name for the user's face.  The file will be inserted
in the buffer as an X-Face field using `x-face-insert' without
argument.  If the file name has no directory component, it should be
found in the directory which is specified by
`x-face-image-file-directory'."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom x-face-insert-query-file-name-when-no-argument nil
  "If it is non-nil, prompt you for a file name unconditionally when
`x-face-insert' is called non-interactively without argument."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-image-file-directory-for-save x-face-image-file-directory
  "Name of the directory where image files should go.  It is referred to
by the command `x-face-save'."
  :version "21.1"
  :type 'directory
  :group 'x-face)

(defcustom x-face-compressor nil
  "Compress method used for saving image files.  This variable can be set
to nil, `gzip' or `bzip2'."
  :version "21.1"
  :type '(radio (const :format "%v " bzip2)
		(const :format "%v " gzip)
		(const :tag "off" nil))
  :group 'x-face)

(defcustom x-face-use-overlay nil
  "If it is non-nil, X-Face will be shown as an overlay with
`before-string' property.  It is useful if you don't want to insert
any extra text in the showing buffer.  However, you should take notice
that some MUAs (e.g. Gnus) might delete such overlays unconditionally.
It will be overridden with t buffer-locally when `x-face-insert' is
done."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-hide-related-fields nil
  "If it is non-nil, not only X-Face fields but also all related fields
\(e.g. X-Face-Version) will be hidden in the showing buffer.  It will
be overridden with nil buffer-locally when `x-face-insert' is done."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defcustom x-face-hidden-properties '(invisible t intangible t)
  "Property list to hide raw X-Face field or its related fields.  It will
be modified buffer-locally in order to transfer an icon image when
`x-face-insert' is done."
  :version "21.1"
  :type 'plist
  :group 'x-face)

(defcustom x-face-auto-image nil
  "If it is non-nil, show X-Face images when `x-face-insert' is done.
It can also be a function, for example, `interactive-p',
`(lambda nil (memq major-mode '(message-mode wl-draft-mode)))', etc."
  :version "21.1"
  :type '(radio (const :format "%t " :tag "OFF" nil)
		(const :format "%t\n" :tag "ON" t)
		(function :format "Function: %v"))
  :group 'x-face)

(defcustom x-face-mouse-face 'highlight
  "Face used to highlight buttons when the mouse moves over it."
  :version "21.1"
  :type 'face
  :group 'x-face)

(defcustom x-face-bbdb-display (and (locate-library "bbdb") t)
  "If it is non-nil, show X-Face images in the BBDB buffer."
  :version "21.1"
  :type 'boolean
  :group 'x-face)

(defvar x-face-ring nil "Cache for face data.")
(defvar x-face-ring-size 32 "Length of the ring used for `x-face-ring'.")

(eval-when-compile
  (defmacro x-face-dark-background-p ()
    "Return t if the selected frame's background is dark."
    '(eq 'dark (frame-parameter (selected-frame) 'background-mode))))

(eval-when-compile
  (defmacro x-face-image-attributes-for-bbdb ()
    "Merge the value of the option `x-face-image-attributes-for-bbdb' into
the copy of `x-face-image-attributes' and return the result."
    '(let* ((bgmode (if (x-face-dark-background-p)
			'dark
		      'light))
	    (basis (copy-sequence (symbol-value 'x-face-image-attributes)))
	    (element (assq bgmode basis))
	    (modifiers (cdr (assq bgmode
				  (symbol-value
				   'x-face-image-attributes-for-bbdb))))
	    attributes)
       (if element
	   (progn
	     (setq attributes (copy-sequence (cdr element)))
	     (while modifiers
	       (setq attributes (plist-put attributes
					   (car modifiers) (cadr modifiers))
		     modifiers (cddr modifiers)))
	     (cons (cons bgmode attributes) (delq element basis)))
	 (cons (cons bgmode modifiers) basis)))))

(put 'x-face-internal-function 'show-sample
     (lambda (&optional bbdb remove-cache)
       "Show an X-Face sample image in a customization buffer.  If BBDB is
non-nil, assign the value of `x-face-image-attributes-for-bbdb' to an
image instead of `x-face-image-attributes'.  If REMOVE-CACHE is non-
nil, reset the value of `x-face-ring' to nil."
       (when (and (display-images-p)
		  (image-type-available-p 'pbm)
		  (eq 'custom-mode major-mode))
	 (save-excursion
	   (let ((inhibit-point-motion-hooks t))
	     (goto-char (point-min))
	     (when (re-search-forward "^<?X Face Image Attributes" nil t)
	       (forward-line 1)
	       (let ((inhibit-read-only t))
		 (delete-region (point)
				(progn
				  (when (looking-at "From:")
				    (forward-line 1)
				    (when (search-forward "X-Face:" nil t)
				      (while (progn
					       (forward-line 1)
					       (memq (char-after)
						     '(?\t ?\ ))))))
				  (skip-chars-forward "\n")
				  (point)))
		 (save-restriction
		   (narrow-to-region (point) (point))
		   (unless
		       (and x-face-default-xbm-file
			    (prog2
				(insert "From: "
					(if user-mail-address
					    (if user-full-name
						(concat
						 user-full-name " <"
						 user-mail-address ">")
					      user-mail-address)
					  "X-Face")
					"\n")
				(condition-case nil
				    (let (x-face-auto-image
					  x-face-hidden-properties)
				      (x-face-insert x-face-default-xbm-file)
				      t)
				  (error
				   (delete-region (point-min) (point-max))
				   nil))
			      (dolist (var '(x-face-use-overlay
					     x-face-hide-related-fields
					     x-face-hidden-properties))
				(kill-local-variable var))))
		     (insert "\
From: X-Face
X-Face: 2i'm.M0UyETCme?'R/?fE}i)R-aY$t;].MSLwmUfB\"^3H+so!vO79{mzviSR4#DM+}\"\"
 ZwMOJ~e&Rr*qL'CrQZo-@jdTL=w{o3Pxu2PY]_qB=w%GLU1S_Pk8HX>4C}W2YTc=V=(~QH[vcm2!]O
 pq&CB^,K:NO/lVI-m&kP;pa&K.Xo)<V4=l<eRW]>X\n"))
		   (when remove-cache
		     (setq x-face-ring nil))
		   (if bbdb
		       (let ((x-face-image-attributes
			      (x-face-image-attributes-for-bbdb))
			     x-face-ring)
			 (x-face-decode-message-header))
		     (let (x-face-ring)
		       (x-face-decode-message-header)))
		   (goto-char (point-min))
		   (forward-line 1)
		   (delete-region (point) (point-max))))))))))

(put 'x-face-internal-function 'image-attributes
     (lambda nil
       "Return a widget type for entering X-Face image attributes."
       (let* ((props (list :ascent :margin :relief :conversion
			   :foreground :background
			   ;; Extra keyword.
			   :scale-factor))
	      (lmax (let ((result 0))
		      (dolist (prop props result)
			(setq result (max result
					  (length (symbol-name prop)))))))
	      (tag "When frame background mode is %v")
	      (left (- (length tag) lmax 4))
	      (selection `(checklist :format
				     ,(format (format "%%%ds%%%%v" (- left 2))
					      "")))
	      prop)
	 (while (setq prop (pop props))
	   (setq selection
		 (nconc
		  selection
		  `((list :format "%v" :inline t
			  (const :format "%v " :value ,prop)
			  (sexp :format
				,(format
				  (format
				   "%%%ds%%-%ds"
				   (- lmax (length (symbol-name prop)))
				   (if props left 0))
				  "" "%v")))))))
	 `(list (cons :format "%v" (sexp :format ,tag) ,selection)
		(cons :format "%v" (sexp :format ,tag) ,selection)))))

(eval-when-compile
  (defmacro x-face-cleanup-plist (plist)
    "Remove properties with the value nil in PLIST."
    `(let ((plist ,plist)
	   property value rest)
       (while (prog1
		  (setq property (pop plist))
		(when (setq value (pop plist))
		  (setq rest (nconc rest (list property value))))))
       rest)))

(put 'x-face-internal-function 'cleanup-attributes
     (lambda (attributes)
       "Remove properties with the value nil in each element of ATTRIBUTES."
       (dolist (element attributes attributes)
	 (setcdr element (x-face-cleanup-plist (cdr element))))))

(defcustom x-face-image-attributes
  '((light :ascent 80 :foreground "#000000")
    (dark :ascent 80 :foreground "#000000" :background "#ffffff"))
  "List of image attributes to assign to the X-Face images.  Each element
consists of a value of the frame background mode (`light' or `dark')
and pairs of a keyword and a value.  This allows the extra keyword
`:scale-factor' used to scale down or scale up the X-Face images.
The value for the keyword `:scale-factor' should be a positive number
or nil (treated as 1)."
  :version "21.1"
  :type (funcall (get 'x-face-internal-function 'image-attributes))
  :get (lambda (symbol)
	 (prog1
	     (let* ((value (default-value symbol))
		    (first (assq (or (frame-parameter (selected-frame)
						      'background-mode)
				     'light)
				 value)))
	       (funcall (get 'x-face-internal-function 'cleanup-attributes)
			(cons first (delq first (copy-sequence value)))))
	   (funcall (get 'x-face-internal-function 'show-sample))))
  :set (lambda (symbol value)
	 (prog1
	     (custom-set-default
	      symbol
	      (funcall (get 'x-face-internal-function 'cleanup-attributes)
		       value))
	   (funcall (get 'x-face-internal-function 'show-sample) nil t)))
  :group 'x-face)

(defcustom x-face-image-attributes-for-bbdb
  '((light :ascent center :foreground "#002000" :background "#fffacd"
	   :relief 2 :scale-factor 0.5)
    (dark :ascent center :foreground "#002000" :background "#fffacd"
	  :relief 2 :scale-factor 0.5))
  "List of attributes to modify the value of `x-face-image-attributes'
which is used to show X-Face images on the BBDB buffer.  The value
form is the same as `x-face-image-attributes'."
  :version "21.1"
  :type (funcall (get 'x-face-internal-function 'image-attributes))
  :get (lambda (symbol)
	 (prog1
	     (let* ((value (default-value symbol))
		    (first (assq (or (frame-parameter (selected-frame)
						      'background-mode)
				     'light)
				 value)))
	       (cons first (delq first (copy-sequence value))))
	   (funcall (get 'x-face-internal-function 'show-sample) t)))
  :set (lambda (symbol value)
	 (prog1
	     (custom-set-default symbol value)
	   (funcall (get 'x-face-internal-function 'show-sample) t t)))
  :group 'x-face)

(defcustom uncompface-program "uncompface"
  "Program used to decode X-Face."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom compface-program "compface"
  "Program used to encode X-Face."
  :version "21.1"
  :type 'string
  :group 'x-face)

(defcustom x-face-pnmscale-program "pnmscale"
  "The \"pnmscale\" executable which comes from the netpbm package.  It
will be used to scale down or scale up X-Face images if the value for
the keyword `:scale-factor' in the option `x-face-image-attributes' or
`x-face-image-attributes-for-bbdb' is neither nil nor number 1.  For
quickly scaling, you can also use \"pnmscalefixed\"."
  :version "21.1"
  :type '(radio (const :format "%v " nil)
		(string))
  :set (lambda (symbol value)
	 (custom-set-default
	  symbol
	  (and (not noninteractive)
	       (stringp value)
	       (with-temp-buffer
		 (set-buffer-multibyte nil)
		 (insert "P1\n1 1\n0\n")
		 (condition-case nil
		     (let ((coding-system-for-read 'binary)
			   (coding-system-for-write 'binary))
		       (call-process-region (point-min) (point-max)
					    value t '(t nil) nil "2")
		       (goto-char (point-min))
		       (let (case-fold-search)
			 (re-search-forward "\
^P5[[:blank:]\n\r]+2[[:blank:]\n\r]+2[[:blank:]\n\r]+" nil t)))
		   (error nil)))
	       value)))
  :group 'x-face)

(defcustom x-face-pgmtoppm-program "pgmtoppm"
  "The \"pgmtoppm\" executable which comes from the netpbm package.
It will be used to colorize the scaled X-Face images."
  :version "21.1"
  :type '(radio (const :format "%v " nil)
		(string))
  :set (lambda (symbol value)
	 (custom-set-default
	  symbol
	  (and (not noninteractive)
	       (stringp value)
	       (with-temp-buffer
		 (set-buffer-multibyte nil)
		 (insert "P1\n2 1\n0 1\n")
		 (condition-case nil
		     (let ((coding-system-for-read 'binary)
			   (coding-system-for-write 'binary))
		       (call-process-region (point-min) (point-max)
					    value t '(t nil) nil
					    "#000000-#ffffff")
		       (goto-char (point-min))
		       (let (case-fold-search)
			 (re-search-forward "\
^P6[[:blank:]\n\r]+2[[:blank:]\n\r]+1[[:blank:]\n\r]+\
\[[:digit:]]+[[:blank:]\n\r]+"
					    nil t)))
		   (error nil)))
	       value)))
  :group 'x-face)

(defvar x-face-field-icon
  (if (image-type-available-p 'xpm)
      (create-image "/* XPM */
static char * x_face_field_icon_xpm[] = {
\"16 14 6 1\",
\"     c None\",
\".    c #000000 s outline\",
\"X    c #ff0000\",
\"o    c #00ff00\",
\"O    c #ffff00\",
\"+    c #00ffff\",
\"                \",
\"......       ...\",
\".XXXXX.     .o. \",
\" .XXXXX.   .o.  \",
\"  .XXXXX. .o.   \",
\"   .XXXXX.o.    \",
\"    .XXX.O.     \",
\"     .X.OOO.    \",
\"    .+.OOOOO.   \",
\"   .+. .OOOOO.  \",
\"  .+.   .OOOOO. \",
\" .+.     .OOOOO.\",
\"...       ......\",
\"                \"};
"
		    'xpm t :ascent 'center)
    (create-image "P1
16 14
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 1 1 1 1 1 0 0 0 0 0 0 0 1 0 0
0 0 1 1 1 1 1 0 0 0 0 0 1 0 0 0
0 0 0 1 1 1 1 1 0 0 0 1 0 0 0 0
0 0 0 0 1 1 1 1 1 0 1 0 0 0 0 0
0 0 0 0 0 1 1 1 0 1 0 0 0 0 0 0
0 0 0 0 0 0 1 0 1 1 1 0 0 0 0 0
0 0 0 0 0 1 0 1 1 1 1 1 0 0 0 0
0 0 0 0 1 0 0 0 1 1 1 1 1 0 0 0
0 0 0 1 0 0 0 0 0 1 1 1 1 1 0 0
0 0 1 0 0 0 0 0 0 0 1 1 1 1 1 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
"
		  'pbm t :ascent 'center :mask 'heuristic))
  "Icon image used to iconified raw X-Face field.  It can be `pbm', `xbm'
or `xpm' image format, or the value nil not to iconify the field.")

(defvar x-face-enable-cache t
  "*Say whether images will be cached in `x-face-ring'.")

(autoload 'ring-elements "ring")
(autoload 'ring-insert "ring")

(defvar x-face-show-buffer nil
  "Internal variable used to say where the X-Face images should be
displayed.")
(defvar x-face-working-buffer " *x-face-working*"
  "Temporary buffer for the internal use.")

(defconst x-face-most-positive-fixnum (eval '(lsh -1 -1))
  "Maximum integer for this Emacs.")

(eval-when-compile
  ;; Avoid byte-compile warnings.
  (autoload 'cmail-folder-buffer "cmail-misc")
  (autoload 'cmail-get-page-number-from-summary "cmail-misc")
  (autoload 'cmail-n-page "cmail-misc")
  (autoload 'cmail-show-contents "cmail")
  (autoload 'gnus-summary-select-article "gnus-sum")
  (autoload 'mew-buffer-message "mew")
  (autoload 'mew-summary-display "mew-summary")
  (autoload 'mh-get-msg-num "mh-utils")
  (autoload 'mh-show-msg "mh-utils")
  (autoload 'vm-follow-summary-cursor "vm-motion")
  (autoload 'wl-message-get-original-buffer "wl-message")
  (autoload 'wl-summary-set-message-buffer-or-redisplay "wl-summary")

  (defsubst x-face-buffer-live-p (buffer)
    "Say whether BUFFER is alive or not."
    (and buffer
	 (get-buffer buffer)
	 (buffer-name (get-buffer buffer))))

  (defsubst x-face-narrow-to-head ()
    "Narrow the buffer to the head of the message.  Point is left at the
beginning of the narrowed-to region."
    (narrow-to-region
     (goto-char (point-min))
     (if (let (case-fold-search)
	   (re-search-forward (concat "^$\\|^"
				      (regexp-quote mail-header-separator)
				      "$")
			      nil t))
	 (match-beginning 0)
       (point-max)))
    (goto-char (point-min)))

  (defmacro x-face-set-properties (start end properties &optional mark)
    "If `x-face-use-overlay' is nil, set PROPERTIES and MARK to the text
from START to END, otherwise use overlay instead of text property."
    `(let ((properties ,properties))
       (when properties
	 (let ((start ,start)
	       (end ,end))
	   (if x-face-use-overlay
	       (let ((overlay (make-overlay start end)))
		 (overlay-put overlay 'evaporate t)
		 (overlay-put overlay 'priority x-face-most-positive-fixnum)
		 ,(when mark
		    `(overlay-put overlay ,mark t))
		 (while properties
		   (overlay-put overlay (pop properties) (pop properties))))
	     ,(when mark
		`(put-text-property start end ,mark t))
	     (let ((s1 (1+ start))
		   (e1 (1- end)))
	       (cond ((< s1 e1)
		      (put-text-property start s1 'front-sticky nil)
		      (put-text-property start e1 'rear-nonsticky nil)
		      (put-text-property s1 end 'front-sticky t)
		      (put-text-property e1 end 'rear-nonsticky t))
		     ((eq s1 e1)
		      (add-text-properties start s1
					   '(front-sticky nil
							  rear-nonsticky nil))
		      (add-text-properties e1 end
					   '(front-sticky t
							  rear-nonsticky t)))
		     ((eq start e1)
		      (add-text-properties start end
					   '(front-sticky nil
							  rear-nonsticky t)))))
	     (add-text-properties start end properties))))))

  (defsubst x-face-remove-x-face-images ()
    "Remove all X-Face images.  The buffer is expected to be narrowed to
just the header of the message."
    (dolist (overlay (overlays-in (point-min) (point-max)))
      (when (overlay-get overlay 'x-face-image)
	(delete-overlay overlay)))
    (let ((inhibit-point-motion-hooks t)
	  (end (point-min))
	  start regions)
      (while (and end
		  (setq start (text-property-any end (point-max)
						 'x-face-image t)))
	(push (cons start
		    (setq end (text-property-not-all start (point-max)
						     'x-face-image t)))
	      regions))
      (let ((mod (buffer-modified-p))
	    (inhibit-read-only t))
	(dolist (region regions)
	  (delete-region (car region) (or (cdr region) (point-max))))
	(set-buffer-modified-p mod))))

  (defsubst x-face-expose-hidden-text ()
    "Expose raw X-Face and its related fields.  The buffer is expected to
be narrowed to just the header of the message."
    (dolist (overlay (overlays-in (point-min) (point-max)))
      (when (overlay-get overlay 'x-face-invisible)
	(delete-overlay overlay)))
    (let ((props (append x-face-hidden-properties '(x-face-invisible t)))
	  (mod (buffer-modified-p))
	  (inhibit-read-only t)
	  (inhibit-point-motion-hooks t)
	  (end (point-min))
	  start)
      (while (and end
		  (setq start (text-property-any end (point-max)
						 'x-face-invisible t)))
	(setq end (text-property-not-all start (point-max)
					 'x-face-invisible t))
	(remove-text-properties start (or end (point-max)) props))
      (set-buffer-modified-p mod)))

  (defmacro x-face-mirror-table ()
    "Return a bitwise symmetrical table for 8-bit data."
    (let ((table (make-vector 256 nil))
	  (i 0)
	  j k l)
      (while (< i 256)
	(setq j i
	      k 0
	      l 8)
	(while (> l 0)
	  (setq k (lsh k -1)
		l (1- l))
	  (if (>= (setq j (lsh j 1)) 256)
	      (setq j (logand 255 j)
		    k (logior 128 k))))
	(aset table i k)
	(setq i (1+ i)))
      table)))

(defconst x-face-mirror (x-face-mirror-table)
  "Bitwise symmetrical table for 8-bit data.")

(defun x-face-possibly-change-buffer ()
  "Possibly change the buffer to where the raw X-Face fields can be
found.  You can redefine it for customization.  The value of the
variable `x-face-show-buffer' can be modified if it is needed."
  (cond ((and (eq 'rmail-summary-mode major-mode)
	      (not (featurep 'rmail-mime)))
	 (set-buffer (symbol-value 'rmail-buffer))
	 (x-face-possibly-change-buffer))
	((and (eq 'rmail-mode major-mode)
	      (not (featurep 'rmail-mime)))
	 (let ((buffer (get-buffer-create x-face-working-buffer)))
	   (save-excursion
	     (goto-char (point-min))
	     (save-restriction
	       (widen)
	       (when (search-backward "\n\n" nil t)
		 (skip-chars-backward "\n")
		 (forward-char))
	       (copy-to-buffer buffer (point)
			       (if (search-backward "\n\n" nil t)
				   (match-end 0)
				 (point-min)))))
	   (set-buffer buffer)))))

(defun x-face-possibly-change-buffer-for-saving ()
  "Possibly change the buffer to where the raw X-Face fields can be
found.  This function will be used for saving X-Face fields.  You can
redefine it for customization."
  (cond ((eq 'cmail-summary-mode major-mode)
	 (let ((page (cmail-get-page-number-from-summary))
	       (buffer (get-buffer-create x-face-working-buffer)))
	   (unless (eq page (symbol-value '*cmail-current-page))
	     (cmail-show-contents page))
	   (set-buffer (cmail-folder-buffer
			(symbol-value 'cmail-current-folder)))
	   (cmail-n-page page)
	   (copy-to-buffer buffer (point) (if (search-forward "\n\n" nil t)
					      (1+ (match-beginning 0))
					    (point-max)))
	   (set-buffer buffer)))
	((eq 'gnus-summary-mode major-mode)
	 (gnus-summary-select-article)
	 (set-buffer (symbol-value 'gnus-original-article-buffer)))
	((and (boundp 'gnus-article-buffer)
	      (eq (current-buffer)
		  (get-buffer (symbol-value 'gnus-article-buffer)))
	      (x-face-buffer-live-p (symbol-value
				     'gnus-original-article-buffer)))
	 (set-buffer (symbol-value 'gnus-original-article-buffer)))
	((memq major-mode '(mew-summary-mode mew-virtual-mode))
	 (mew-summary-display nil)
	 (set-buffer (mew-buffer-message)))
	((eq 'mh-folder-mode major-mode)
	 (let* ((num (mh-get-msg-num t))
		(buffer (symbol-value 'mh-show-buffer))
		(folder (symbol-value 'mh-current-folder))
		(raw-buffer (concat "article-" (if (bufferp folder)
						   (buffer-name folder)
						 folder))))
	   (if (and (x-face-buffer-live-p buffer)
		    (string-match (format "/%d$" num)
				  (buffer-file-name (get-buffer buffer))))
	       (set-buffer (if (x-face-buffer-live-p raw-buffer)
			       raw-buffer
			     buffer))
	     (mh-show-msg num)
	     (if (x-face-buffer-live-p raw-buffer)
		 (set-buffer raw-buffer)
	       (current-buffer)))))
	((eq 'rmail-summary-mode major-mode)
	 (set-buffer (symbol-value 'rmail-buffer))
	 (x-face-possibly-change-buffer-for-saving))
	((eq 'rmail-mode major-mode)
	 (let ((buffer (get-buffer-create x-face-working-buffer)))
	   (save-excursion
	     (goto-char (point-min))
	     (save-restriction
	       (widen)
	       (when (search-backward "\n\n" nil t)
		 (skip-chars-backward "\n")
		 (forward-char))
	       (copy-to-buffer buffer (point)
			       (if (search-backward "\n\n" nil t)
				   (match-end 0)
				 (point-min)))))
	   (set-buffer buffer)))
	((eq 'vm-summary-mode major-mode)
	 (vm-follow-summary-cursor)
	 (if (x-face-buffer-live-p (symbol-value 'vm-mail-buffer))
	     (set-buffer (symbol-value 'vm-mail-buffer))
	   (error "Folder buffer has been killed")))
	((eq 'wl-summary-mode major-mode)
	 (wl-summary-set-message-buffer-or-redisplay)
	 (set-buffer (wl-message-get-original-buffer)))))

;;;###autoload
(defun x-face-create-image (data &rest props)
  "Create a PBM image from DATA.  DATA is an encoded X-Face string which
may have or may not have a \"X-Face:\" header, spaces, tabs or
newlines.  The rest PROPS are additional image attributes to assign to
the image.  The value of the variable `x-face-image-attributes' will
be used to the default image attributes.

Here are some examples of how to use this function:

;; Insert an image at point.
\(insert-image (x-face-create-image DATA :KEYWORD VALUE ...))

;; Create an image without any properties.
\(let (x-face-image-attributes)
  (x-face-create-image DATA))

;; Create a scaled image.
\(x-face-create-image DATA :scale-factor 0.707)

;; Insert an image as an overlay.
\(let ((overlay (make-overlay (point) (1+ (point))))
      (image (x-face-create-image DATA)))
  (overlay-put overlay 'evaporate t)
  (overlay-put overlay 'before-string (propertize \" \" 'display image)))
"
  (if (stringp data)
      (progn
	(setq data (split-string data))
	(if (string-equal "X-Face:" (car data))
	    (setq data (apply 'concat (cdr data)))
	  (setq data (apply 'concat data))))
    ;; Return an empty X-Face by default.
    (setq data ",\\m{?h\\)X"))
  (let* ((params (copy-sequence (cdr (assq (if (x-face-dark-background-p)
					       'dark
					     'light)
					   x-face-image-attributes))))
	 (scale (or (plist-get props :scale-factor)
		    (plist-get params :scale-factor)
		    1))
	 bg fg should-colorize siblings image
	 (coding-system-for-read 'binary)
	 (coding-system-for-write 'binary))
    (while props
      (setq params (plist-put params (pop props) (pop props))))
    (setq params (x-face-cleanup-plist (plist-put params :scale-factor nil))
	  bg (or (plist-get params :background)
		 (face-background 'default)
		 "#ffffff")
	  fg (or (plist-get params :foreground)
		 (face-foreground 'default)
		 "#000000")
	  should-colorize
	  (and x-face-pgmtoppm-program
	       (not (and (string-match
			  "^white$\\|^gr[ae]y100$\\|^#\\(?:fff\\)+$" bg)
			 (string-match
			  "^black$\\|^gr[ae]y0$\\|^#\\(?:000\\)+$" fg)))))
    (plist-put params :background bg)
    (plist-put params :foreground fg)
    (when x-face-enable-cache
      (if x-face-ring
	  (setq siblings (cdr (assoc data (ring-elements x-face-ring)))
		image (cdr (assoc scale siblings)))
	(setq x-face-ring (make-ring x-face-ring-size))))
    (unless image
      ;; Create an image.
      (with-temp-buffer
	(set-buffer-multibyte nil)
	(setq case-fold-search t)
	(insert data)
	(call-process-region (point-min) (point-max)
			     uncompface-program t '(t nil))
	(goto-char (point-min))
	(while (search-forward "0x" nil t)
	  (replace-match ""))
	(goto-char (point-min))
	(while (re-search-forward "[^[:xdigit:]]+" nil t)
	  (replace-match ""))
	(goto-char (point-min))
	(insert "\"P4\n48 48\n")
	(while (not (eobp))
	  (insert "\\x")
	  (forward-char 2))
	(insert "\"")
	(goto-char (point-min))
	(insert (prog1
		    (string-as-unibyte (read (current-buffer)))
		  (erase-buffer)))
	(when (and x-face-pnmscale-program
		   (/= 1 scale))
	  (call-process-region (point-min) (point-max)
			       x-face-pnmscale-program t '(t nil) nil
			       (number-to-string scale))
	  (goto-char (point-min))
	  (if (re-search-forward "^P[56]" nil t)
	      (delete-region (point-min) (match-beginning 0))
	    (error "Scaling image failed"))
	  (when should-colorize
	    (call-process-region (point-min) (point-max)
				 x-face-pgmtoppm-program t '(t nil) nil
				 (concat fg "-" bg))
	    (goto-char (point-min))
	    (if (re-search-forward "^P6" nil t)
		(unless (eq (match-beginning 0) (point-min))
		  (message "%s" "Maybe colorize image failed")
		  (delete-region (point-min) (match-beginning 0)))
	      (error "Colorize image totally failed"))))
	(setq image (list 'image :type 'pbm :data (buffer-string)))
	(when x-face-enable-cache
	  (if siblings
	      (push (cons scale image) siblings)
	    (ring-insert x-face-ring (list data (cons scale image)))))))
    (append image params)))

(defun x-face-gray-x-faces-to-image (x-faces)
  "Convert gray X-Faces to a PGM image."
  (let* ((params (copy-sequence (cdr (assq (if (x-face-dark-background-p)
					       'dark
					     'light)
					   x-face-image-attributes))))
	 (size (or (plist-get params :scale-factor)
		   1))
	 siblings image bit-arrays)
    (dolist (face (prog1
		      (nreverse x-faces)
		    (setq x-faces nil)))
      (push (apply 'concat (split-string face)) x-faces))
    (when x-face-enable-cache
      (if x-face-ring
	  (setq siblings (cdr (assoc x-faces (ring-elements x-face-ring)))
		image (cdr (assoc size siblings)))
	(setq x-face-ring (make-ring x-face-ring-size))))
    (unless image
      (let ((faces x-faces)
	    x-face-ring x-face-image-attributes)
	(while faces
	  (push
	   ;; Make a bool vector.
	   (read (concat "#&2304"
			 (prin1-to-string
			  (substring
			   (plist-get (cdr (x-face-create-image (pop faces)))
				      :data)
			   9))))
	   bit-arrays)))
      (let* ((ix 0)
	     iy
	     (iz 14)
	     (depth (length x-faces))
	     (scale (/ 255 (1- (expt 2 depth))))
	     (picture (make-vector 2304 nil))
	     pixel)
	(while (< ix 2304)
	  (setq iy 0
		pixel 0)
	  (while (< iy depth)
	    (setq pixel (if (aref (nth iy bit-arrays) (+ ix iz -7))
			    (+ pixel pixel scale)
			  (* pixel 2))
		  iy (1+ iy)))
	  (aset picture ix pixel)
	  (setq ix (1+ ix)
		iz (if (zerop iz)
		       14
		     (- iz 2))))
	(if (and x-face-pnmscale-program
		 (/= 1 size))
	    (let ((coding-system-for-read 'binary)
		  (coding-system-for-write 'binary)
		  default-enable-multibyte-characters)
	      (with-temp-buffer
		;; We should use `concat' to convert a vector to a string.
		(insert (string-as-unibyte (concat "P5\n48 48\n255\n"
						   picture)))
		(call-process-region (point-min) (point-max)
				     x-face-pnmscale-program t '(t nil) nil
				     (number-to-string size))
		(setq image (create-image (buffer-string) 'pbm t))))
	  (setq image (create-image (string-as-unibyte
				     (concat "P5\n48 48\n255\n" picture))
				    'pbm t)))
	(when x-face-enable-cache
	  (if siblings
	      (push (cons size image) siblings)
	    (ring-insert x-face-ring (list x-faces (cons size image)))))))
    (plist-put params :background nil)
    (plist-put params :foreground nil)
    (plist-put params :scale-factor nil)
    (append image (x-face-cleanup-plist params))))

(eval-when-compile
  (defvar last))

;;;###autoload
(defun x-face-decode-message-header (&optional beg end buffer)
  "Display X-Faces in the current message.  Optional BEG and END are
ignored.  If optional BUFFER is specified, it is assumed that the raw
X-Face fields can be found in the BUFFER.  BUFFER can also be a
function which is similar to `x-face-possibly-change-buffer'.  This
requires support for images in your Emacs and the external program
`uncompface'.  (Optional BEG and END are no more than placeholders to
ensure the compatibility with the old programs.)"
  (when (and (display-images-p)
	     (image-type-available-p 'pbm))
    (let ((inhibit-point-motion-hooks t)
	  start gray depth images)
      (save-excursion
	;; Remove all existing X-Face images.
	(save-restriction
	  (x-face-narrow-to-head)
	  (x-face-remove-x-face-images))
	(setq x-face-show-buffer (current-buffer))
	;; Change buffer to where the raw X-Face fields can be found.
	(cond ((functionp buffer)
	       (funcall buffer))
	      ((x-face-buffer-live-p buffer)
	       (set-buffer buffer))
	      (t
	       (x-face-possibly-change-buffer)))
	;; Extract X-Faces in header.
	(save-restriction
	  (x-face-narrow-to-head)
	  (let ((case-fold-search t)
		x-faces)
	    (while (progn
		     (while (not (or (eobp)
				     (looking-at
				      "X-Face\\(-[[:digit:]]+\\)?:")))
		       (forward-line 1))
		     (not (eobp)))
	      (setq start (match-end 0))
	      (if (match-beginning 1)
		  ;; There are gray X-Face data.
		  (setq gray t
			depth (- (string-to-number (match-string 1))))
		(setq depth 0))
	      (while (progn
		       (forward-line 1)
		       (memq (char-after) '(?\t ?\ ))))
	      (push (cons depth
			  (buffer-substring-no-properties start (point)))
		    x-faces))
	    (if gray
		(setq images
		      (list (x-face-gray-x-faces-to-image
			     (dolist (element
				      (prog1
					  (sort x-faces 'car-less-than-car)
					(setq x-faces nil))
				      x-faces)
			       (push (cdr element) x-faces)))))
	      (while x-faces
		(push (x-face-create-image (cdr (pop x-faces))) images))))))
      (when (x-face-buffer-live-p x-face-working-buffer)
	(kill-buffer x-face-working-buffer))
      (when images
	;; Insert X-Face images.
	(save-excursion
	  (set-buffer x-face-show-buffer)
	  (let ((case-fold-search t)
		(inhibit-read-only t)
		(mod (buffer-modified-p)))
	    (save-restriction
	      (x-face-narrow-to-head)
	      (let ((from (if (re-search-forward "^From:" nil t)
			      (prog1
				  (match-end 0)
				(goto-char (point-min)))
			    (point))))
		;; Hide raw fields.
		(let ((icon (plist-get x-face-hidden-properties 'display)))
		  (when icon
		    (setq icon (copy-sequence icon))
		    (let ((type (plist-get (cdr icon) :type))
			  (dark (x-face-dark-background-p))
			  (bg (plist-get (cdar images) :background))
			  (fg (plist-get (cdar images) :foreground)))
		      (cond ((eq type 'xpm)
			     (plist-put (cdr icon) :color-symbols
					(list (cons "outline"
						    (if dark
							bg
						      fg)))))
			    ((memq type '(pbm xbm))
			     (plist-put (cdr icon) :foreground
					(if dark
					    bg
					  fg))))))
		  (let ((regexp (if x-face-hide-related-fields
				    "X-Face.*: ?"
				  "X-Face\\(-[[:digit:]]+\\)?: ?")))
		    (while (progn
			     (while (not (or (eobp) (looking-at regexp)))
			       (forward-line 1))
			     (not (eobp)))
		      (setq start (if icon
				      (match-end 0)
				    (point)))
		      (while (progn
			       (forward-line 1)
			       (memq (char-after) '(?\t ?\ ))))
		      (if icon
			  (x-face-set-properties start (1- (point))
						 (plist-put
						  (copy-sequence
						   x-face-hidden-properties)
						  'display icon)
						 'x-face-invisible)
			(if (and (boundp 'gnus-article-buffer)
				 (eq (get-buffer (symbol-value
						  'gnus-article-buffer))
				     (current-buffer)))
			    ;; We should not put the hidden properties at
			    ;; the last newline since it will be left when
			    ;; sorting header is done.
			    (x-face-set-properties start (1- (point))
						   x-face-hidden-properties
						   'x-face-invisible)
			  (x-face-set-properties start (point)
						 x-face-hidden-properties
						 'x-face-invisible))))))
		(goto-char from)))
	    (if x-face-use-overlay
		(let ((overlay (make-overlay (point) (1+ (point))))
		      (rest ""))
		  (setq images (nreverse images))
		  (while images
		    ;; There is a face `default' used to hide an
		    ;; underline text behind an image.
		    (setq rest (concat (propertize " " 'display
						   (pop images) 'face 'default)
				       rest)))
		  (overlay-put overlay 'before-string
			       (if (bolp)
				   (concat rest "\n")
				 rest))
		  (overlay-put overlay 'x-face-image t)
		  (overlay-put overlay 'priority x-face-most-positive-fixnum)
		  (overlay-put overlay 'evaporate t))
	      (while images
		(setq start (point))
		(when (prog1
			  (when (bolp)
			    (insert "From:")
			    t)
			(insert-image (pop images)))
		  (insert " X-Face\n"))
		(put-text-property start (point) 'x-face-image t)))
	    (set-buffer-modified-p mod)))))
    (when (and (boundp 'gnus-article-buffer)
	       (eq (current-buffer) (symbol-value 'gnus-article-buffer)))
      (set 'last t))))

;;;###autoload
(defun x-face-show (&optional arg)
  "Toggle showing X-Face images.  With ARG, turn showing on if and only
if ARG is positive."
  (interactive "P")
  (if (if (numberp arg)
	  (> arg 0)
	(not (or (and (numberp arg) (< arg 0))
		 ;; There is no From field in the buffer.
		 (get-text-property (point-min) 'x-face-image)
		 (save-excursion
		   (save-restriction
		     (x-face-narrow-to-head)
		     (or (let ((overlays (overlays-in (point-min)
						      (point-max))))
			   (while (and overlays
				       (not (overlay-get (car overlays)
							 'x-face-image)))
			     (setq overlays (cdr overlays)))
			   overlays)
			 (text-property-any (point-min) (point-max)
					    'x-face-image t)))))))
      (x-face-decode-message-header)
    (x-face-remove-x-face-images)
    (x-face-expose-hidden-text)))

;;;###autoload
(defun x-face-turn-off ()
  "Remove X-face images from the buffer."
  (x-face-show -1))

(eval-when-compile
  (defmacro x-face-insert-error (error-symbol &rest args)
    "Signal ERROR-SYMBOL with concatenated ARGS."
    (list 'let* (list (list 'args (cons 'list args))
		      '(rest (if (interactive-p)
				 (pop args)
			       "[x-face-insert]")))
	  (list 'signal error-symbol
		'(list (dolist (arg args rest)
			 (setq rest (concat rest " " arg))))))))

(defun x-face-read-xbm-completion-handler (operation &rest args)
  "Internal function used to handle completion for XBM files."
  (let ((inhibit-file-name-handlers
	 (cons 'x-face-read-xbm-completion-handler inhibit-file-name-handlers))
	(inhibit-file-name-operation operation)
	(partial (car args))
	(dir (cadr args))
	(case-fold-search t)
	candidates)
    (if (eq operation 'file-name-completion)
	(try-completion
	 partial
	 (dolist (name (file-name-all-completions partial dir) candidates)
	   (setq candidates (nconc candidates (list (list name))))))
      (if (eq operation 'file-name-all-completions)
	  (dolist (name (file-name-all-completions partial dir) candidates)
	    (when (string-match "\\(?:/\\|\\.xbm\\(?:\\.bz2\\|\\.gz\\)?\\)\\'"
				name)
	      (setq candidates (nconc candidates (list name)))))
	(apply operation args)))))

;;;###autoload
(defun x-face-read-xbm-file-name (prompt &optional dir default-filename
					 mustmatch initial)
  "Read XBM file name.  This function is equivalent to `read-file-name',
except that it limits to XBM files."
  (let ((file-name-handler-alist
	 (cons (cons "" 'x-face-read-xbm-completion-handler)
	       file-name-handler-alist)))
    (read-file-name prompt dir default-filename mustmatch initial)))

;;;###autoload
(defun x-face-insert (&optional xbm-file keep-existing-fields)
  "Insert XBM-FILE as an X-Face field.  If XBM-FILE is omitted, the value
of `x-face-default-xbm-file' will be used for it.  If the contents of
a file do not look like the XBM format, they will be regarded as pre-
encoded data.  If optional KEEP-EXISTING-FIELDS is non-nil, existing
X-Face fields will not be removed.  This requires the external program
`compface'.  It will override the value of the variable
`x-face-use-overlay' to t buffer-locally."
  (interactive "*i\nP")
  (unless (and (stringp xbm-file) (> (length xbm-file) 0))
    (let (dir default)
      (when (and (stringp x-face-default-xbm-file)
		 (> (length x-face-default-xbm-file) 0))
	(setq dir (file-name-directory x-face-default-xbm-file)
	      default
	      (if (and dir
		       (let ((inode1 (nth 10 (file-attributes dir)))
			     (inode2 (nth 10 (file-attributes
					      x-face-image-file-directory))))
			 (and inode1 inode2 (equal inode1 inode2))))
		  (file-name-nondirectory x-face-default-xbm-file)
		x-face-default-xbm-file)))
      (if (or (interactive-p)
	      x-face-insert-query-file-name-when-no-argument)
	  (setq xbm-file
		(x-face-read-xbm-file-name
		 (concat "XBM file"
			 (if default
			     (concat " (default " default "): ")
			   ": "))
		 (or dir x-face-image-file-directory) default t))
	(when default
	  (setq xbm-file x-face-default-xbm-file))))
    (unless (and (stringp xbm-file) (> (length xbm-file) 0))
      (x-face-insert-error 'file-error "No xbm file is specified")))
  (setq xbm-file (expand-file-name xbm-file x-face-image-file-directory))
  (when (file-directory-p xbm-file)
    (x-face-insert-error 'file-error xbm-file "is a directory"))
  (or (file-exists-p xbm-file)
      (and (file-exists-p (concat xbm-file ".bz2"))
	   (setq xbm-file (concat xbm-file ".bz2")))
      (and (file-exists-p (concat xbm-file ".gz"))
	   (setq xbm-file (concat xbm-file ".gz")))
      (x-face-insert-error 'file-error "Cannot open XBM file:" xbm-file))
  (let ((x-face
	 (with-temp-buffer
	   (set-buffer-multibyte nil)
	   (if (and (string-match "\\.bz2$\\|\\.gz$" xbm-file)
		    (not auto-compression-mode))
	       (progn
		 (auto-compression-mode 1)
		 (unwind-protect
		     (insert-file-contents xbm-file)
		   (auto-compression-mode -1)))
	     (if auto-image-file-mode
		 (progn
		   (auto-image-file-mode -1)
		   (insert-file-contents xbm-file)
		   (auto-image-file-mode 1))
	       (insert-file-contents xbm-file)))
	   (setq case-fold-search t)
	   (cond ((and
		   (search-forward "width" nil t)
		   (condition-case nil
		       (eq 48 (read (current-buffer)))
		     (error nil))
		   (goto-char (point-min))
		   (search-forward "height" nil t)
		   (condition-case nil
		       (eq 48 (read (current-buffer)))
		     (error nil))
		   (search-forward "{" nil t)
		   (progn
		     (delete-region (point-min) (point))
		     (while (search-forward "0x" nil t)
		       (replace-match ""))
		     (goto-char (point-min))
		     (while (re-search-forward "[^[:xdigit:]]+" nil t)
		       (replace-match ""))
		     (eq 576 (buffer-size))))
		  (goto-char (point-min))
		  (insert "\"")
		  (while (not (eobp))
		    (insert "\\x")
		    (forward-char 2))
		  (insert "\"")
		  (goto-char (point-min))
		  (dolist (byte (prog1
				    (append (string-as-unibyte
					     (read (current-buffer))) nil)
				  (erase-buffer)))
		    (insert (format "%02x" (aref x-face-mirror byte))))
		  (call-process-region (point-min) (point-max)
				       compface-program t '(t nil))
		  (goto-char (point-min))
		  (while (search-forward "\r" nil t)
		    (replace-match ""))
		  (goto-char (point-min))
		  (skip-chars-forward "\t\n ")
		  (delete-region (point-min) (point))
		  (goto-char (point-max))
		  (skip-chars-backward "\t\n ")
		  (delete-region (point) (point-max)))
		 ((and (> (buffer-size) 0)
		       (progn
			 (goto-char (point-min))
			 (skip-chars-forward "\t\n\v\f\r -~")
			 (eobp)))
		  (goto-char (point-min))
		  (if (search-forward "\nX-Face" nil t)
		      ;; There may be multiple X-Faces.
		      (progn
			(goto-char (point-min))
			(skip-chars-forward "\t\n\r ")
			(delete-region (point-min) (point))
			(while (re-search-forward "[[:blank:]\r]+$" nil t)
			  (replace-match ""))
			(goto-char (point-max))
			(skip-chars-backward "\t\n ")
			(delete-region (point) (point-max)))
		    (while (re-search-forward "[[:blank:]\n\v\f\r]+" nil t)
		      (replace-match ""))
		    (goto-char (point-min))
		    (move-to-column 70)
		    (while (not (eobp))
		      (insert "\n ")
		      (move-to-column 78))))
		 (t
		  (x-face-insert-error 'file-error
				       "Unrecognized file format:" xbm-file)))
	   (buffer-string))))
    (set (make-local-variable 'x-face-use-overlay) t)
    (set (make-local-variable 'x-face-hide-related-fields) nil)
    (let ((props (default-value 'x-face-hidden-properties)))
      (set (make-local-variable 'x-face-hidden-properties)
	   (when (and x-face-field-icon
		      (plist-get props 'invisible))
	     (list 'display (copy-sequence x-face-field-icon)
		   'intangible (plist-get props 'intangible)))))
    (save-excursion
      (save-restriction
	(let ((inhibit-point-motion-hooks t))
	  (x-face-narrow-to-head)
	  (let (x-faces)
	    (let ((case-fold-search t))
	      (while (progn
		       (while (not (or (eobp)
				       (looking-at "X-Face:")))
			 (forward-line 1))
		       (not (eobp)))
		(push (cons (point)
			    (progn
			      (while (progn
				       (forward-line 1)
				       (memq (char-after) '(?\t ?\ ))))
			      (point)))
		      x-faces))
	      (goto-char (or (cdar x-faces) (point-max)))
	      (let ((inhibit-read-only t)
		    (keymap (current-local-map))
		    props start)
		(setq keymap (if keymap
				 (copy-keymap keymap)
			       (make-sparse-keymap)))
		(define-key keymap [mouse-2] 'x-face-show)
		(setq props (list 'local-map keymap
				  'mouse-face x-face-mouse-face
				  'help-echo
				  "mouse-2 toggles X-Face visibility"))
		(unless (bolp) (insert "\n"))
		(save-restriction
		  (narrow-to-region (point) (point))
		  (insert "X-Face: ")
		  (setq start (point))
		  (insert x-face "\n")
		  (goto-char start)
		  (forward-line 1)
		  (while (not (eobp))
		    (while (memq (char-after) '(?\t ?\ ))
		      (forward-line 1))
		    (x-face-set-properties start (1- (point)) props)
		    (while (prog1
			       (not (or (eobp)
					(and (looking-at
					      "X-Face:[[:blank:]\n]+")
					     (setq start (match-end 0)))))
			     (forward-line 1)))))
		(unless keep-existing-fields
		  (while (setq x-face (pop x-faces))
		    (delete-region (car x-face) (cdr x-face)))))))))))
  (when (if (functionp x-face-auto-image)
	    (funcall x-face-auto-image)
	  x-face-auto-image)
    (x-face-show 1)))

;;;###autoload
(defun x-face-save (&optional filename)
  "Save X-Face fields to XBM files.  This requires the external program
`uncompface'.  It currently does not work with the forwarded messages
except for Mew."
  (interactive)
  (let ((case-fold-search t)
	x-faces from)
    (save-excursion
      (x-face-possibly-change-buffer-for-saving)
      (save-restriction
	(x-face-narrow-to-head)
	(let ((inhibit-point-motion-hooks t)
	      start)
	  (while (progn
		   (while (not (or (eobp) (looking-at "X-Face:")))
		     (forward-line 1))
		   (not (eobp)))
	    (setq start (match-end 0))
	    (while (progn
		     (forward-line 1)
		     (memq (char-after) '(?\t ?\ ))))
	    (setq x-faces (nconc x-faces
				 (list (buffer-substring-no-properties
					start (1- (point))))))))
	(when (setq from (mail-fetch-field "from"))
	  (setq from (cadr (mail-extract-address-components from))))))
    (when (x-face-buffer-live-p x-face-working-buffer)
      (kill-buffer x-face-working-buffer))
    (if x-faces
	(let* ((minibuffer-setup-hook (append minibuffer-setup-hook
					      '(end-of-line)))
	       (default (concat
			 (if from from "x-face") ".xbm"
			 (cdr (assq x-face-compressor '((bzip2 . ".bz2")
							(gzip . ".gz"))))))
	       (filename
		(or filename
		    (x-face-read-xbm-file-name
		     "XBM file: "
		     x-face-image-file-directory-for-save
		     default nil default)))
	       (directory (or (file-name-directory filename)
			      x-face-image-file-directory-for-save))
	       (filename (file-name-nondirectory filename))
	       (basename
		(if (string-match
		     "\\(?:-[[:digit:]]+\\)?\\.xbm\\(?:\\.bz2\\|\\.gz\\)?$"
		     filename)
		    (substring filename 0 (match-beginning 0))
		  filename))
	       (basename-re (concat "^" (regexp-quote basename)))
	       (num 0)
	       (log-buffer (get-buffer-create "*X-Face-Save-Log*"))
	       nform suffix compr)
	  (if (file-directory-p directory)
	      (let ((files (directory-files directory nil basename-re t)))
		(when files
		  (let ((re (concat basename-re "-\\([[:digit:]]+\\)"))
			file)
		    (while (setq file (pop files))
		      (setq num (max num (if (string-match re file)
					     (1+ (string-to-number
						  (match-string 1 file)))
					   1)))))))
	    (make-directory directory t))
	  (with-current-buffer log-buffer
	    (setq case-fold-search nil)
	    (goto-char (point-max))
	    (unless (bolp) (insert "\n"))
	    (buffer-disable-undo))
	  (setq nform (concat "%s-%0" (number-to-string
				       (max 2 (length (number-to-string num))))
			      "d"))
	  (when (string-match "\\(\\.xbm\\)\\(\\.bz2\\|\\.gz\\)?$" filename)
	    (setq suffix (match-string 1 filename)
		  compr (match-string 2 filename)))
	  (while x-faces
	    (setq filename (if (zerop num)
			       basename
			     (format nform basename num))
		  num (1+ num))
	    (with-temp-buffer
	      (set-buffer-multibyte nil)
	      (insert (car x-faces))
	      (setq x-faces (cdr x-faces)
		    case-fold-search t)
	      (call-process-region (point-min) (point-max) uncompface-program
				   t '(t nil))
	      (goto-char (point-min))
	      (while (search-forward "0x" nil t)
		(replace-match ""))
	      (goto-char (point-min))
	      (while (re-search-forward "[^[:xdigit:]]+" nil t)
		(replace-match ""))
	      (goto-char (point-min))
	      (insert "\"")
	      (while (not (eobp))
		(insert "\\x")
		(forward-char 2))
	      (insert "\"")
	      (goto-char (point-min))
	      (dolist (byte
		       (prog1
			   (append (string-as-unibyte (read (current-buffer)))
				   nil)
			 (erase-buffer)
			 (insert "#define " filename "_width 48\n"
				 "#define " filename "_height 48\n"
				 "static char " filename "_bits[] = {\n ")))
		(insert (format "0x%02X," (aref x-face-mirror byte)))
		(when (eq (current-column) 76)
		  (insert "\n ")))
	      (delete-char -1)
	      (insert "};\n")
	      (setq filename (expand-file-name (concat filename suffix compr)
					       directory))
	      (if (and compr
		       (require 'jka-compr)
		       (not auto-compression-mode))
		  (progn
		    (auto-compression-mode 1)
		    (unwind-protect
			(write-region (point-min) (point-max) filename)
		      (auto-compression-mode -1)))
		(write-region (point-min) (point-max) filename))
	      (when (and x-faces (interactive-p))
		(sit-for 0.5))
	      (with-current-buffer log-buffer
		(insert (format-time-string "%Y-%m-%d %T ")
			(abbreviate-file-name filename)
			"\n")
		(bury-buffer log-buffer)))))
      (message "No X-Face fields found"))))

(eval-and-compile
  (autoload 'bbdb-current-record "bbdb-com")
  (autoload 'bbdb-record-getprop "bbdb")
  (autoload 'bbdb-record-name "bbdb"))

(defun x-face-energize-bbdb-buffer (&optional all silent)
  "Show X-Face images in the BBDB buffer.  If optional ALL is not
specified, it deals with only the current record."
  (when x-face-bbdb-display
    (save-excursion
      (set-buffer (symbol-value 'bbdb-buffer-name))
      (save-restriction
	(unless all
	  (beginning-of-line)
	  (narrow-to-region (progn
			      (while (and (memq (char-after)
						'(nil ?\t ?\n ?\ ))
					  (zerop (forward-line -1))))
			      (point))
			    (progn
			      (while (progn
				       (forward-line 1)
				       (memq (char-after) '(?\t ?\ ))))
			      (point))))
	(x-face-remove-x-face-images)
	(when (and x-face-bbdb-display
		   (display-images-p)
		   (image-type-available-p 'pbm))
	  (let ((buffer (get-buffer-create " *x-face-energize-bbdb*"))
		(x-face-use-overlay t)
		(picon (when x-face-field-icon
			 (let* ((icon (copy-sequence x-face-field-icon))
				(type (plist-get (cdr icon) :type))
				(bgmode (if (x-face-dark-background-p)
					    'dark
					  'light))
				(params
				 (x-face-cleanup-plist
				  (cdr (assq bgmode x-face-image-attributes))))
				(bg (plist-get params :background))
				(fg (plist-get params :foreground)))
			   (unless (stringp bg)
			     (unless (stringp
				      (setq bg (face-background 'default)))
			       (setq bg "#ffffff")))
			   (unless (stringp fg)
			     (unless (stringp
				      (setq fg (face-foreground 'default)))
			       (setq fg "#000000")))
			   (cond ((eq type 'xpm)
				  (plist-put (cdr icon) :color-symbols
					     (list (cons "outline"
							 (if (eq 'dark bgmode)
							     bg
							   fg)))))
				 ((memq type '(pbm xbm))
				  (plist-put (cdr icon) :foreground
					     (if (eq 'dark bgmode)
						 bg
					       fg))))
			   icon)))
		(x-face-image-attributes (x-face-image-attributes-for-bbdb))
		;; Space with the face `default' used to stop leaking
		;; of neighboring faces.
		(space (propertize " " 'face 'default))
		record faces msg nfaces start end from to icon)
	    (goto-char (point-min))
	    (unwind-protect
		(while (not (eobp))
		  (when all
		    ;; Go to the beginning or the next recored.
		    (while (and (looking-at "[[:blank:]]\\|[[:blank:]]*$")
				(zerop (forward-line 1)))))
		  (when (and (not (eobp))
			     (setq record (bbdb-current-record))
			     (setq faces (bbdb-record-getprop record 'face)))
		    (when (and all
			       (not silent))
		      (setq msg (format "Extracting X-Face(s) for %s..."
					(bbdb-record-name record)))
		      (message "%s" msg))
		    (setq faces (split-string faces)
			  nfaces (length faces))
		    (with-current-buffer buffer
		      (erase-buffer)
		      (insert "From: X\n")
		      (dolist (face faces)
			(insert "X-Face: " face "\n"))
		      (x-face-decode-message-header)
		      (setq faces (overlay-properties (car (overlays-at 6)))))
		    (plist-put faces 'before-string
			       (concat (plist-get faces 'before-string)
				       space))
		    (x-face-set-properties (point) (1+ (point)) faces)
		    (forward-line 1)
		    (when picon
		      ;; Iconify the raw X-Face records.
		      (while (not (or (when (looking-at "[[:blank:]]+face:")
					(goto-char (match-end 0)))
				      (looking-at "[^[:blank:]\n]*$")
				      (eobp)))
			(forward-line 1))
		      (unless (or (bolp) (eobp))
			(setq start (point))
			(while (progn
				 (forward-line 1)
				 ;; There might not be a blank line between
				 ;; records when BBDB v2.33 and later is
				 ;; running.
				 (not (or (looking-at "\
\[^[:blank:]\n]\\|[[:blank:]]+[^[:blank:]\n:]+:\
\\(?:[[:blank:]]\\|[[:blank:]]*$\\)")
					  (eobp)))))
			(setq end (point))
			(goto-char start)
			(while (re-search-forward "[^[:blank:]\n]+" end t)
			  (setq from (match-beginning 0)
				to (match-end 0)
				nfaces (1- nfaces))
			  (if (zerop nfaces)
			      ;; Use the parent icon directly.
			      (setq icon picon)
			    (when (and (looking-at "[[:blank:]\n]+")
				       (< (match-end 0) end))
			      (setq to (match-end 0)
				    icon (copy-sequence picon))))
			  (x-face-set-properties from to
						 (list 'display icon
						       'intangible t
						       'x-face-image t))))))
		  (forward-line 1))
	      (when msg
		(message "%sdone" msg))
	      (kill-buffer buffer))))))))

(let (current-load-list)
  (defadvice bbdb-display-records-1
    (around show-x-face-images-in-the-bbdb-buffer activate compile)
    "Advised by X-Face-E21.  Show X-Faces images in the BBDB buffer."
    (let ((silent (or (and (boundp 'bbdb-gag-messages)
			   (symbol-value 'bbdb-gag-messages))
		      (and (boundp 'bbdb-silent-running)
			   (symbol-value 'bbdb-silent-running)))))
      (unless silent
	(message "Formatting..."))
      (let ((bbdb-silent-running t)
	    (bbdb-list-hook bbdb-list-hook))
	(remove-hook 'bbdb-list-hook 'x-face-energize-bbdb-buffer)
	ad-do-it)
      (x-face-energize-bbdb-buffer 'all silent)
      (unless silent
	(message "Formatting...done")))))

(add-hook 'bbdb-list-hook 'x-face-energize-bbdb-buffer)

(provide 'x-face-e21)

(eval-when-compile
  (when noninteractive
    (message "%s" "\r
  ++ Do `isearch' for the string \"STRONGLY\" in x-face-e21.el\r
  ++ and see around there for your happy networking.\r
  ++ (It is not an error message, never mind :-)\r
")))

;;; x-face-e21.el ends here
